#!/usr/bin/perl -w

# ====================================================================
# The Vovida Software License, Version 1.0 
# 
# Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. The names "VOCAL", "Vovida Open Communication Application Library",
#    and "Vovida Open Communication Application Library (VOCAL)" must
#    not be used to endorse or promote products derived from this
#    software without prior written permission. For written
#    permission, please contact vocal\@vovida.org.
# 
# 4. Products derived from this software may not be called "VOCAL", nor
#    may "VOCAL" appear in their name, without prior written
#    permission of Vovida Networks, Inc.
# 
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
# NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
# NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
# IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
# 
# ====================================================================
# 
# This software consists of voluntary contributions made by Vovida
# Networks, Inc. and many individuals on behalf of Vovida Networks,
# Inc.  For more information on Vovida Networks, Inc., please see
# <http://www.vovida.org/>.

use strict;
use Getopt::Long;

use Socket;
use Tk;
#! Set tk_strict_Motif to get rid of silly highlighting
use Telephony::SipMessage qw ( _new_sdp_headers );
use Telephony::SipServer;
use FancyBalloon;

my $program_name   = 'Tricorder';
my %settings = (arrow_start_vposition    => 15,
		window_default_font      => '-B&H-Lucida-Medium-R-Normal-Sans-14-140-75-75-P-81-ISO8859-1',
		

		canvas_default_font      => '-Adobe-Courier-Medium-R-Normal--12-120-75-75-M-70-ISO8859-1',
		break_font               => '-Adobe-Courier-Medium--Normal--24-120-75-75-M-70-ISO8859-1',
		
		canvas_width             => 500, #minimum; adjusted below
		canvas_height            => 400, #minimum; adjusted below
		
		
		space_between_arrows     => 40,
		
		server_start_hposition   => 50,
		server_vposition         => 20,
		space_between_servers    => 50,
		server_box_hsize         => 170, #! Later, scale to text of name
		server_box_vsize         => 50,

		break_width              => 10,
		break_offset_space       => 15
	       	);

#Change these variables to tweak Tricorder's appearance.
my %colors =   (INVITE     => 'brown4', 
		ACK        => 'green3',
		OPTIONS    => 'orange',
		BYE        => 'darkgoldenrod4',
		REGISTER   => 'purple',

		'1xx'      => 'black',
		'200'      => 'forest green',
		'3xx'      => 'blue',
		'4xx'      => 'red',
		'5xx'      => 'dark red',

		sip_header_name    => 'black',
		sip_header_value   => 'blue',
		sdp_header_name    => 'dark slate grey',
		sdp_header_value   => 'green',

		window_background  => 'grey',
		window_foreground  => 'black',
		canvas_background  => 'light yellow',
		canvas_foreground  => 'black',
		server_box_fill    => 'white',
		server_box_outline => 'black',

		break              => 'black'
		);

#parameters that control the Postscript printing functionality
#These are just defaults; the user's preferred settings are loaded
#from the user's .tricorderrc if it exists.
my %snapshot_preferences = (file           => 'call_flow.eps',
			    colormode      => 'mono',
			    scaling_factor => 1,
			    rotate         => 1, #landscape orientation
			    page_length    => 1100, #pixels
			    page_width     => 800, #pixels
			    );

my $ipgrab_command = 'ipgrab -t -l -n -f ';
my $window;
my $canvas;
my @sip_messages;
#Side-effected by &draw_arrow
my ($arrow_vposition, $arrow_number);
#Side-effected by &draw_server_box
my $server_box_hposition;
&set_initial_positions;

#Use a hash to optimize membership lookups, since both the source and the
#destination of every SIP message will be checked for membership in the
#set of known servers--with a list, that lookup would become expensive.
my %server_hash;
#The list is there only to preserve the order of the servers.  Its information
#is redundant with that in the hash.  That's O.K.; they hold references, not
#values, so (almost) nothing is lost
#This approach is ever-so-slightly faster (albeit less elegant) than using
#a pseudo-hash, because it saves the array lookup in the implicit
#$pseudo_hash[0]{'key'} done at access time.
my @server_list;

#Process options and, in their absence, set default values.

my $debug_postscript = 0;
my $debug_snapshots  = 0;
my $ueber_debug_snapshots = 1;

my ($help, $log_file, $use_ip_numbering_scheme, $silent, $debug, $debug_sip_messages, $batch_process, $cisco, $heuristic, $tee, $ipgrab_expression, $paginate_snapshots);
GetOptions ('--help'                     => \$help,
	    '--log-file=s'               => \$log_file,
	    '--silent'                   => \$silent,
	    '--debug'                    => \$debug,
	    '--debug-sip-messages'       => \$debug_sip_messages,
	    '--batch-process=s'          => \$batch_process,
	    '--cisco'                    => \$cisco,
	    '--heuristic=s'              => \$heuristic,
	    '--tee=s'                    => \$tee,
	    '--ipgrab-expression=s'      => \$ipgrab_expression,
	    '--paginate-snapshots:s'     => \$paginate_snapshots
	    );

$cisco = 1;

if (defined $help)
{
    &usage();
}
$debug = 1 if $debug_sip_messages; #--debug-sip-messages turns on normal --debug
if (defined $debug)
{
    if (defined $silent)
    {
	&usage ('You may not specify --silent and --debug at once.');
    }
    else
    {
	require Data::Dumper;
        Data::Dumper->import();
    }
}
if (defined $log_file)
{
    if (defined $ipgrab_expression)
    {
	&usage ("It makes no sense to specify --log-file and --ipgrab-expression at once,\n"
	     . "since using a pre-made log file implies that ipgrab will not"
	     . "be run interactively.");
    }
    elsif (defined $tee)
    {
	&usage ("It makes no sense to specify --log-file and --tee at once,\n"
	     . "since using a pre-made log file implies that you already have\n"
	     . "a log, hence you don't need to create one with tee.");
    }
}

$ipgrab_command .= (defined $ipgrab_expression) ? $ipgrab_expression : 'port 5060';
print "\$ipgrab_command is $ipgrab_command\n" if $debug;

if (defined $log_file || defined $batch_process && (! $silent))
{
    warn <<fnord;
Warning:  For the log file to work, it must have been produced by
ipgrab using the -t, -n, and -l options.  $program_name
currently does not understand more verbose ipgrab output.

fnord
}
else #default:  read from a running ipgrab
{
    $log_file = "$ipgrab_command|";
}
if (defined $heuristic)
{
    unless ($heuristic eq 'assume-server' ||
	    $heuristic eq 'use-vovida-ip-numbering-scheme' ||
	    $heuristic eq 'use-sip-hints')
    {
	&usage ("You entered an invalid argument to --use-heuristic.\n"
	      . "Please use one of the arguments described below.");
    }
}
else # $heuristic is not defined
{
    $heuristic = 'assume-server'; #default:  Treat everything as a server.
}
if (defined $tee)
{
    open (TEE, "<$tee") #!sysopen for security!
	or warn "Could not open $tee to collect ipgrab output:  $!";
}
if (defined $paginate_snapshots)
{
    if (! $paginate_snapshots)
    {
	#the default if no argument to --paginate-snapshots is given
	$paginate_snapshots = 'dumb'; 
    }
    elsif ($paginate_snapshots ne 'smart' && $paginate_snapshots ne 'dumb')
    {
	usage ("You entered an invalid argument to --paginate-snapshots.\n"
	       . "Please use one of the arguments described below.");
    }
}

#Read preferences from the .tricorderrc file, if it exists.
my $HOME = $ENV{HOME} || '.';
if (-r "$HOME/.tricorderrc")
{
    #Read in and eval the saved value of %snapshot_preferences, which
    #was written by Data::Dumper.
    print "Found a .tricorderrc; reading preferences...\n" if $debug;

    my $preferences;

    open RC_FILE, "$HOME/.tricorderrc"
	or die "Race condition:  $HOME/.tricorderrc was readable a moment ago, but now it is not:  $!";
    while (<RC_FILE>) { $preferences .= $_; }
    eval $preferences;

    close RC_FILE
	or warn "Could not close $HOME/.tricorderrc:  $!";

    print "Done reading preferences.\n" if $debug;
}

#! untested code for batch mode, not yet added to the product spec
#  if (defined $batch_process)
#  {
#      my $i = 1;
#      foreach my $filename (split /\s+/, $batch_process)
#      {
#  	print "Working on log file $filename...\n" unless $silent;
#  	&parse_log_file ($filename);
#  	&draw_canvas(); #off-screen; don't bother with the rest of the GUI
#  	my $filename_sans_directory;
#  	if ($filename =~ m|/([^/]+)$|)
#  	{
#  	    $filename_sans_directory = $1;
#  	}
#  	else
#  	{
#  	    $filename_sans_directory = $filename;
#  	}

#  	#Take reasonable pains to the file name unique.
#  	&take_snapshot ("tricorder_batch_output--$$--$filename_sans_directory--$i.eps");
#  	$window->withdraw();
#  	@sip_messages = (); #Clear it out for the next round; reclaim memory.
#  	@server_list   = []; #ibid.
#  	%server_hash   = (); #op. cit.

#  	$i++;
#      }
#      exit; #success!
#  }

print "Opening log file $log_file...\n" unless $silent;
open LOG_FILE, $log_file #!Ack!  Insecure!  Use sysopen() instead.
    or die "Could not open ipgrab log $log_file:  $!";
print "Starting graphical user interface...\n" unless $silent;
&create_gui();

#Start the Tk event loop.
#Invoke the parsing callback whenever there is input to read.
$window->fileevent (\*LOG_FILE, 'readable', \&parse);
#The rest of the time, process GUI events.
MainLoop();

my $i = 0;    
my @sip_message;
my @sdp_message;
my $starting_new_sip_message = 0;
my $starting_new_sdp_message = 0;
my $in_sip_message = 0;
my $in_sdp_message = 0;

#Figure out where on the canvas to start drawing arrows and servers,
#and set global variables appropriately.
sub set_initial_positions
{
    $arrow_vposition = $settings{server_box_vsize}
                     + $settings{server_vposition}
                     + $settings{arrow_start_vposition};
    $arrow_number = 1;
    $server_box_hposition = $settings{server_start_hposition};
}

#Parse a single line and accumulate lines into messages.  Once the end of a
#message is reached, parse it as a whole and construct a new SipMessage,
#then draw an arrow for it on the canvas.  Discover and draw new servers
#in a similarly dynamic fashion.
#! Uses blocking <>; later, replace with non-blocking calls?
sub parse
{
    if (! ($_ = <LOG_FILE>))
    {
	if ($!)
	{
	    die "I/O error:  $!";
	}
	else
	{
	    #Try to construct a SIP message from the latest batch of
	    #accumulated output.  This may fail, but it will succeed
	    #in two important cases:
	    #1.  The SIP message is complete except for the trailing line
	    #of dashes.  ipgrab seems to omit that line in most cases,
	    #so it is important to make this attempt to make sure the
	    #last message appears.
	    #2.  ipgrab output was cut off (e.g. by the user pressing
	    #Ctrl-C) while listing SIP headers.  As a result, the pop-up header
	    #list will be incomplete, but at least the request itself
	    #and the extant headers will be displayed correctly.
	    my $message = Telephony::SipMessage->new ($debug_sip_messages,
						      \@sip_message);
	    print "Reached end of SIP messages.\n"
		. "Got final SIP message:\n" . Data::Dumper->Dump ([ $message ]) if $debug_sip_messages;
	    close LOG_FILE;
	}
    }
    else
    {
	print "Got a line of available input:\n\t$_" if $debug_sip_messages;
	print TEE if $tee;
	if (m/^-+$/) #In ipgrab output, lines of dashes separate messages
	{
	    if ($in_sip_message)
	    {
		my $message = Telephony::SipMessage->new ($debug_sip_messages,
							  \@sip_message);
		$in_sip_message = 0; @sip_message = qw();
		push @sip_messages, $message;
		$i++;
		#An unblessed reference means the message was incomplete,
		#which means we hit the end of the file.
		last if (ref $message ne 'Telephony::SipMessage');
		&do_the_cisco_massage ($message) if defined $cisco;
		if ($debug_sip_messages)
		{
		    print "Constructed SipMessage\n"
			. Data::Dumper->Dump ([ $message ])
			    . "\n\n";
		}
	    
		&handle_new_server ($message, 'source');
		&handle_new_server ($message, 'destination');
		&draw_arrow ($message);
		&resize_canvas();
	    }
	    elsif ($in_sdp_message)
	    {
		#Attach to last SIP message
		($sip_messages[$i - 1])->{sdp_headers} = Telephony::SipMessage->_new_sdp_headers (\@sdp_message);
		$in_sdp_message = 0; @sdp_message = qw();
	    }
	    elsif ($starting_new_sip_message)
	    {
		$starting_new_sip_message = 0;
		$in_sip_message = 1;
	    }
	    elsif ($starting_new_sdp_message)
	    {
		$starting_new_sdp_message = 0;
		$in_sdp_message =1;
	    }
	}
	elsif (m/SIP Headers/)
	{
	    $starting_new_sip_message = 1;
	}
	elsif (m/SDP Headers/)
	{
	    $starting_new_sdp_message = 1;
	}
	else #It's part of a message proper.
	{
	    if ($in_sip_message)
	    {
		push @sip_message, $_;
	    }
	    elsif ($in_sdp_message)
	    {
		push @sdp_message, $_;
	    }
	    else
	    {
		warn "Parser found unexpected data between messages:\n$_";
	    }
	}
    }
}

sub handle_new_server
{
    my ($message, $endpoint) = @_;

    #Add the source and destination servers if they do not already
    #exist.
    my $ip          = $message->{"${endpoint}_ip"} || '';
    my $port        = $message->{"${endpoint}_port"};
    my $ip_and_port = $ip . ':' . $port;

    if ($ip && ! exists ($server_hash{$ip_and_port})
	    && &looks_like_a_server ($message, $endpoint))
    {
	my $server = Telephony::SipServer->new ($ip, $port);
	$server_hash{$ip_and_port} = $server;
	push @server_list, $server;
	print "Found new server (IP $ip, port $port):\n"
	    . Data::Dumper->Dump ([ $server ]) if $debug;
	&draw_server_box ($server);
	&resize_canvas();
    }
}

# Massage the source and destination IP to account for
# the fact that Cisco IP phones use high local ports,
# rather than 5060.  Otherwise the phone appears as
# two different boxes, since a box is determined by
# the combination of IP and port.  (This is necessary
# to allow for the discrimination of multiple servers
# running on a single machine.)
sub do_the_cisco_massage
{
    my $cutoff = 9999;
    my $message = $_[0];

    foreach my $fieldname ('source_port', 'destination_port')
    {
	if ($message->{$fieldname} > $cutoff)
	{
	    $message->{$fieldname} = 5060;
	    print "\&do_the_cisco_massage changed port $message->{$fieldname} to 5060\n" 
		. "for message " . Data::Dumper->Dump ([ $message ]) . "\n" if $debug_sip_messages;
	}
    }
}

#Resize the canvas to make it large enough for all that has been drawn on it.
sub resize_canvas
{
    my ($x1, $y1, $x2, $y2) = my @bounding_coordinates = $canvas->bbox ('all');
    print "Resizing canvas:  ($x1, $y1), ($x2, $y2)\n" if $debug;
    $canvas->configure (scrollregion => \@bounding_coordinates);
    &draw_lines_under_server_boxes ($y2); #bottom of the canvas
}

sub read_message_text
{
    my @message_text; my $i = 0;
    while (<LOG_FILE>)
    {
	return \@message_text if m/^-+$/;
	$message_text[$i++] = $_;
    }
    return \@message_text; #Return incomplete text at end of log;
                           #Maybe someone can make something of it.
}

#Apply the user's choice of heuristics to the recipient or sender of a SIP message
#to guess whether it is coming from a server or just from some random
#endpoint.
sub looks_like_a_server
{
    my ($message, $endpoint) = @_;

    if (ref $message ne 'Telephony::SipMessage')
    {
	warn "\&looks_like_a_server got an argument of type " . ref $message
	    . "\n, which is not a SipMessage.  Skipping...";
	return;
    }
    
    #default:  Treat everything as a server.
    return 1 if ($heuristic eq 'assume-server');

    #another option:  Discriminate between servers and endpoints based on
    #their IPs

    if ($heuristic eq 'use-vovida-ip-numbering-scheme')
    {
	my $ip = $message->{"${endpoint}_ip"};

	#Check to see if its IP fits the Vovida SIP server numbering scheme.
	$ip =~ m|\.(.+)$|m;  #Look at the
	my $last_octet = $1; #last octet only.
	if ($last_octet == 180 || #a UA marshal, by Vovida conventions
	    $last_octet == 200 || #a redirect server, by Vovida conventions
	    $last_octet == 110 || #a 5300 marshal, by Vovida conventions
	    $last_octet == 210)   #a 5300, by Vovida conventions
	{
	    return 1;
	}
	else
	{
	    return 0;
	}
    }
    #another option:  Be really clever and look at the actual SIP messages.
    elsif ($heuristic eq 'use-sip-hints')
    {
	my $ip = $message->{"${endpoint}_ip"};
	if ($message->{sip_uri} =~ /@(\.+?):/)
	{
	    #works for both hostnames and IPs...
	    my $ip_from_URI = inet_ntoa (gethostbyname ($1));

	    if ($endpoint eq 'destination')
	    {
		#Check to see if the IP from the SIP-URI matches the
		#destination IP; if so, it's an endpoint, not a server.
		if ($ip_from_URI eq $ip) { return 0; }
		else { return 1; }
	    }
	    else #by deduction, ($endpoint eq $SOURCE)
	    {
		#Check to see if the host in the SIP-URI matches the From header;
		#if so, it's an endpoint, not a server
		my $ip_from_From = $message->get_sip_header ('From');
		if ($ip_from_From eq $ip) { return 0; }
		else { return 1; }
	    }
	}
	else
	{
	    warn "Got a SIP message with an invalid SIP-URI:  \n"
		. Data::Dumper->Dump ([ $message ]);
	}
    }
    else
    {
	#Default:  Assume it's a server, just to be safe.
	return 1;
    }
}

sub create_gui
{
    print "Entered create_gui\n" if $debug;
    $window = MainWindow->new (background => $colors{window_background},
			       foreground => $colors{window_foreground}
			       #-font       => $settings{window_default_font}
			       );
    $window->title ("Tricorder                  " . &get_quip());
    $canvas = $window->Scrolled ('Canvas',
				 -scrollbars   => 'ose',
				 width         => $settings{canvas_width},
				 height        => $settings{canvas_height},
				 relief        => 'flat',
				 background    => $colors{canvas_background}
				 #foreground   => $colors{canvas_foreground}
				 #! font       => $settings{canvas_default_font}
				 );
    $canvas->pack (fill => 'both', expand => 1);
    my $snapshot_button = $window->Button (-text    => 'Take a Snapshot!',
					-command => sub
					            {
							#!&remove_server_boxes();
							&take_snapshot();
							#!&restore_server_boxes();
							#!&draw_lines_under_server_boxes();
						    }
					);
    $snapshot_button->pack();
    my $snapshot_configure_button = $window->Button (-text    => 'Snapshot Properties...',
						     -command => \&configure_snapshot_properties);
    $snapshot_configure_button->pack();
    my $clear_button = $window->Button (-text    => 'Clear Screen',
					-command => \&clear_screen);
    $clear_button->pack();
    my $insert_break_button = $window->Button (-text    => 'Insert Break',
					       -command => \&draw_insert_break_dialog);
    $insert_break_button->pack();
}


#Better than making them local, or passing them around willy-nilly to subs...
my ($least_x, $least_y, $greatest_x, $greatest_y);
my ($page_width, $page_length);
#We control the vertical.  We control the horizontal.
my ($x1, $y1, $x2, $y2);
my $output_filename;
my $file_number;
my $smart;
sub take_snapshot
{
    ($page_width, $page_length) = @snapshot_preferences{'page_width', 'page_length'};
    $output_filename = $snapshot_preferences{file};
    $file_number = 0;
    ($least_x, $least_y, $greatest_x, $greatest_y) = $canvas->bbox ('all');
    if ($debug_snapshots)
    {
	$canvas->createRectangle ($least_x, $least_y, $greatest_x, $greatest_y, outline => 'red');
    }

    print "Canvas dimensions are ($least_x, $least_y), ($greatest_x, $greatest_y)\n" if $ueber_debug_snapshots;

    unless (defined $paginate_snapshots) #Just take one big snapshot.
    {
	&take_snapshot_page ($least_x, $least_y, $greatest_x, $greatest_y, $output_filename);
	return;
    }

    #Divide the canvas into pages of fixed width (configured through the
    #"Snapshot Properties" button).  If --paginate-snapshots was set to smart,
    #use additional formatting heuristics.

    if ($paginate_snapshots eq 'dumb')
    {
	$x1 = $least_x;             #set here as well as in &do_horizontal
	$x2 = $x1 + $page_width;    #so that &crop_canvas will have all the
	                            #variables it needs to work the first
	                            #time it is invoked, before &do_horizontal
	$y1 = $least_y;
	$y2 = $y1 + $page_length;
	
	if ($y2 >= $greatest_y)
	{
	    $y2 = $greatest_y;
	    &do_horizontal(); #one last time...
	    return;
	}
	while (1)
	{
	    &do_horizontal();
	}
	continue
	{
	    $y1 = $y2;
	    $y2 += $page_length;
	    if ($y2 >= $greatest_y)
	    {
		$y2 = $greatest_y;
		&do_horizontal();
		return;
	    }
	}
    }
    else # $paginate_snapshots eq 'smart', by deduction
    {
	$x1 = $least_x;
	$x2 = $greatest_x;
	$y1 = $least_y;
	$y2 = $y1 + $page_length;
	if ($y2 >= $greatest_y)
	{
	    &do_snapshot(); #one last time...
	    return;
	}
	while (1)
	{
	    &adjust_snapshot_boundaries();
	    &do_snapshot();
	}
	continue
	{
	    $y1 = $y2;
	    $y2 += $page_length;
	    if ($y2 >= $greatest_y)
	    {
		$y2 = $greatest_y;
		&do_snapshot();
		return;
	    }
	}
    }
}

sub do_horizontal
{
    if ($x2 >= $greatest_x)
    {
	$x2 = $greatest_x;
	&do_snapshot();
	return;
    }
    while (1) { &do_snapshot(); }
    continue
    {
	$x1 = $x2;
	$x2 += $page_width;
	if ($x2 >= $greatest_x)
	{
	    $x2 = $greatest_x;
	    &do_snapshot();
	    return;
	}
    }
}

#Erase servers and their window dressing from the canvas.
sub remove_server_boxes
{

    $canvas->delete ($canvas->find ('withtag', 'server_box'));
    $canvas->delete ($canvas->find ('withtag', 'vertical_line'));
    $canvas->delete ($canvas->find ('withtag', 'server_label'));

}

#Restore servers and their window dressing to the canvas.
sub restore_server_boxes
{
    map { &draw_server_box ($_); } @server_list;    
}

sub do_snapshot
{
    print "Entering \&do_snapshot, coordinates are ($x1, $y1), ($x2, $y2)\n" if $ueber_debug_snapshots;
    $canvas->createRectangle ($x1, $y1, $x2, $y2, outline => 'blue') if $ueber_debug_snapshots;

    &take_snapshot_page ($x1,
			 $y1,
			 $x2,
			 $y2,
			 &increment_filename ($output_filename));
}

#Push orphaned SIP messages onto the next page.  The idea
#is to keep separate instances of a single request or response
#together when they are propagating through a series of servers.
#For instance, all the 200 messages in a response to a BYE should
#appear on the same page, even if that means leaving some blank space
#on the preceding page.
sub adjust_snapshot_boundaries
{
    my @items = $canvas->find ('enclosed', $x1, $y1, $x2, $y2);
    print "The enclosed test gave coordinates ($x1, $y1), ($x2, $y2)\n" if $ueber_debug_snapshots;

    if ($debug_snapshots)
    {
	print "Adjusting items:\n";
	foreach my $item (@items)
	{
	    print "$item:  " . $canvas->gettags ($item) . "\n";
	}
	print "\n";
    }

    #The items are already ordered from first-drawn (i.e. lowest on the
    #canvas) to last-drawn (i.e. highest on the canvas).
 
    #If the last-drawn item is a label, it must be an orphan to a message
    #arrow on the following page.

    my @filtered_items;
    foreach my $item (@items)
    {
	my $type = $canvas->type ($item);
	print "$item is a $type.  " if $debug_snapshots;
	if ($type ne 'text')
	{
	    #It's either an arrow or a bit of window dressing like a server
            #box's horizontal line.  Either way, it's guaranteed not to be an
	    #orphan.  Include it.
	    print "Keeping it...\n" if $debug_snapshots;
	    push @filtered_items, $item;
	}
	else
	{
	    my $text = $canvas->itemcget ($item, '-text');
	    print ":  '$text'.\n" if $debug_snapshots;
	    #Keep server boxes and labels and the lines drawn under them
	    if (grep /server|line/, $canvas->gettags ($item))
	    {
		push @filtered_items, $item;
		print "It's a server glyph.  Keeping it.\n" if $debug_snapshots;
		next;
	    }
	    $item =~ m|^(\d+)|;
	    if (&in_bbox ("message_arrow_$1", \@items))
	    {
		print "Keeping $item (a label with text $text)." if $debug_snapshots;
		push @filtered_items, $item;
	    }
            #Otherwise leave $item out of the new list, since its arrow is out.
	    elsif ($debug_snapshots)
	    {
		print "Throwing out $item (a label with text $text):"
		    . "\tIts corresponding message arrow was not included in this page.\n";
	    }    
	}
    }

    #Find the type of SIP message label that belongs to the first
    #arrow on the next page (i.e. the first canvas item after the last
    #item on this page)
    my $last_item = $filtered_items[$#filtered_items];
    my $next_label = $canvas->find ('above', $last_item);
    until ($canvas->type ($next_label) eq 'text')
    {
	$next_label = $canvas->find ('above', $next_label);
	if (! defined $next_label)
	{
	    #An undefined label means we've reached the end of the output.
	    ($x1, $y1, $x2, $y2) = $canvas->bbox (@items);
	    last;
	}
	print "...and \$next_label is $next_label\n" if $debug_snapshots;
    }

    if ($debug_snapshots)
    {
	my @tags = $canvas->gettags ($next_label);
	print "The first label on the next page is "
	. Data::Dumper->Dump ([ $next_label ])
	. "\nIts tags are "
	. Data::Dumper->Dump ([ \@tags ]) if $debug_snapshots;
    }

    my $next_label_text = $canvas->itemcget ($next_label, '-text');
    print "\n\tIts text is $next_label_text\n" if $debug_snapshots;
    #Extract the SIP message type itself.
    my $next_SIP_message_type = '';
    if (defined $next_label_text)
    {
	$next_label_text =~ m|\d+\.\s+(.*)|;
	$next_SIP_message_type = $1;
	print "Its message type is $next_SIP_message_type\n\n" if $debug_snapshots;
    }

    #Go backwards through the all the labels on the last page.
    for (my $i = $#filtered_items; $i > 0; $i--)
    {
	my $label = $filtered_items[$i];
	next unless $canvas->type ($label) eq 'text';
	my $label_text = $canvas->itemcget ($label, '-text');
	if (defined $label_text)
	{
	    print "Examining message '$label_text'...\n" if $debug_snapshots;
	    $label_text =~ m|\d+\.\s+(.*)|;
	    my $SIP_message_type = $1;
	    if ($SIP_message_type eq $next_SIP_message_type)
	    {
		print "It's another $next_SIP_message_type.  Push it to the next page.\n" if $debug_snapshots;
		#If it's the same kind of message, it belongs on the next
		#page.  Remove both its label and its arrow from our list
		#of items.
		splice @filtered_items, $i, 2;
	    }
	    else
	    {
		print "It's a different type of message.\nLeave it where it is, and stop looking for orphans.\n\n" if $debug_snapshots;
		last;
	    }
	}
    }

    #Finally, resize the page to include the new set of items, sans orphans.
    #(Don't mess with the x-coordinates; those must remain fixed to ensure
    #that the whole horizontal spread will be squeezed down to fit on one page.)
    ($y1, $y2) = ($canvas->bbox (@filtered_items))[1,3];
    #Fudge the upper bound to include the first SIP message label.
    #Otherwise it will not quite fit (due to a quirk in the bbox method).
    $y1 -= 10 unless $file_number == 0; #Don't try it with the first snapshot!
}

sub in_bbox
{
    my ($arrow_id, @items) = @_;
    foreach my $item (@items)
    {
	print "Looking in " . $canvas->gettags ($item) . "...\n" if $debug_snapshots;
	return 1 if grep /$arrow_id/, $canvas->gettags ($item);
    }
    return 0;
}

#Cut the canvas down so there are no items (servers or messages)
#hanging off its edges.
sub crop_canvas
{
    print "Cropping from ($x1, $y1), ($x2, $y2)...\n" if $debug_snapshots;
    my @items = $canvas->find ('enclosed', $x1, $y1, $x2, $y2);
    unless (defined @items)
    {
	print "Found no enclosed items!\n" if $ueber_debug_snapshots;
	return;
    }
    #Leave $x1 and $y1 intact no matter what.  They are guaranteed to be
    #justified just past the ending position of the last snapshot area
    #(or, in the case of the first snapshot, justified to the top of the page).
    
    #In the case of smart pagination, we know that the x range will include
    #the whole viewable area of the canvas, so we leave $x1 and $x2 alone too. 
    if ($paginate_snapshots eq 'smart')
    {
	$y2 = ($canvas->bbox (@items))[3];
    }
    #In the case of dumb pagination, the x range can vary.
    elsif ($paginate_snapshots eq 'dumb')
    {
	($x2, $y2) = ($canvas->bbox (@items))[2,3];
    }
    
    print "         to   ($x1, $y1), ($x2, $y2)\n\n" if $debug_snapshots;
    return \@items;
}

#Insert a number into a filename, so that the various pages written
#by &take_snapshot_page have unique names.
sub increment_filename
{
    my $filename = $_[0];
    print "Incrementing $filename to " if $debug;

    print "(the number is $file_number; incrementing it to " if $debug;
    $file_number++;
    print "$file_number)... " if $debug;
    #For Windows users' sake, put the number before the first dot,
    #if there is one.
    {    
	last if ($filename =~ s|(\..*)|$file_number$1|);
	$filename .= $file_number;
    }
    print "$filename\n" if $debug;
    return $filename;
}

sub take_snapshot_page
{
    my ($x1, $y1, $x2, $y2, $output_filename, $page_width) = @_;
    print "Saving snapshot for coordinates ($x1, $y1), ($x2, $y2) in $output_filename\n" if $ueber_debug_snapshots;

    my $width  = $x2 - $x1;
    my $height = $y2 - $y1;
    print "Width is $width; height is $height\n" if $debug_snapshots;

    my $postscript = $canvas->postscript (x          => $x1,
					  y          => $y1,
					  width      => $width,
					  height     => $height,
					  colormode  => $snapshot_preferences{colormode},
					  rotate     => $snapshot_preferences{rotate},
					  );
    
    print "Got postscript:\n$postscript" if $debug_postscript;
    if (! open DIAGRAM, ">$output_filename")
    {
	&show_error ("Could not open the Postscript output file $output_filename for writing:  $!");
	return;
    }
    print "Printing Postscript to $output_filename...\n" if $debug_snapshots;
    print DIAGRAM $postscript;
    if (! close DIAGRAM)
    {
	&show_error ("Could not close the Postscript output file $output_filename:  $!" );
	return;
    }
    print "Closed $output_filename.\n" if $debug_snapshots;
}

sub show_error
{
    if (defined $batch_process)
    {
	warn ($_[0]);
	return;
    }

    my $pop_up = MainWindow->Toplevel (background => $colors{window_background},
				       foreground => $colors{window_foreground},
				       takefocus  => 1);
    my $warning   = $pop_up->Label ($_[0]);
    my $OK_button = $pop_up->Button (text    => 'OK',
				     command => sub { $warning->withdraw(); });
    $warning->pack();
    $OK_button->pack();
}

#Create a dialog that lets the user change some snapshot properties
#--specifically, the target filename, the color mode, and the horizontal
#and vertical scaling.
sub configure_snapshot_properties
{
    my ($file, $colormode, $scaling_factor, $page_width, $page_length)
	= @snapshot_preferences{'file', 'colormode', 'scaling_factor', 'page_length', 'page_width'};

    my $left_justify = 30; #pixels
    my $maximum_widget_width   = 600;

    my $dialog = $window->Toplevel (width      => $left_justify * 2
				                + $maximum_widget_width,
				    background => $colors{window_background},
				    foreground => $colors{window_foreground},
				    takefocus  => 1);
    $dialog->title ('Configure');

    my $frame1 = $dialog->Frame (width => $maximum_widget_width);
    my $frame2 = $dialog->Frame (width => $maximum_widget_width);
    my $frame3 = $dialog->Frame (width => $maximum_widget_width);
    my $button_frame = $dialog->Frame (width => $maximum_widget_width);

    my $file_entry = $frame1->Entry();
    $file_entry->insert (0, $file);
    my $file_entry_label = $frame1->Label (text => 'Output to file');

    $colormode = &canonical_to_verbose ($snapshot_preferences{colormode});
    print "Going into snapshot configuration, color mode is $colormode\n" if $debug;
    my $color = $frame2->Radiobutton (variable => \$colormode,
				      value    => 'Color',
				      text     => 'Color');
    my $greyscale = $frame2->Radiobutton (variable => \$colormode,
					  value    => 'Greyscale',
					  text     => 'Greyscale');
    my $mono = $frame2->Radiobutton (variable => \$colormode,
				     value    => 'Black and White',
				     text     => 'Black and White');
    my $scaling_factor_scale = $frame3->Scale (orient       => 'horizontal',
					       from         => .50,
					       to           => 5.00,
					       tickinterval => .50,
					       label        => 'Scaling Factor',
					       length       => $maximum_widget_width,
					       variable     => \$scaling_factor);
    my $page_width_scale = $frame3->Scale (orient       => 'horizontal',
					   from         => 500,
					   to           => 5000,
					   tickinterval => 500,
					   label        => 'Page Width (Pixels)',
					   length       => $maximum_widget_width,
					   variable     => \$page_width);

    my $page_length_scale = $frame3->Scale (orient       => 'horizontal',
					    from         => 500,
					    to           => 10000,
					    tickinterval => 1000,
					    label        => 'Page Height (Pixels)',
					    length       => $maximum_widget_width,
					    variable     => \$page_length);

    my $OK_button = $button_frame->Button (text    => 'OK',
				     command => sub
				                {
						    @snapshot_preferences{'file', 'colormode', 'scaling_factor', 'page_width', 'page_length'} = ($file_entry->get(), &verbose_to_canonical ($colormode), $scaling_factor, $page_width, $page_length);
						    print "New snapshot prefernces:\n" . Data::Dumper->Dump ([ \%snapshot_preferences ]) if $debug;
						    $dialog->withdraw();
						    &save_snapshot_preferences();
						}
				     );
    my $Cancel_button = $button_frame->Button (text    => 'Cancel',
					 command => sub
					            {
							$dialog->withdraw();
						    }
					 );
    $frame1->pack();
    $frame2->pack();
    $frame3->pack();
    $button_frame->pack();
    
    #! Later replace this with a Tk::FileSelect
    $file_entry_label->pack (side => 'left');
    $file_entry->pack       (side => 'left', fill => 'x');

    $color->pack            (side => 'left');
    $greyscale->pack        (side => 'left');
    $mono->pack             (side => 'left');

    $scaling_factor_scale->pack (fill => 'x');
    $page_width_scale->pack     (fill => 'x');
    $page_length_scale->pack    (fill => 'x');
    
    $OK_button->pack        (side => 'left');
    $Cancel_button->pack    (side => 'right');
}

sub clear_screen
{
    @sip_messages = [];
    @sdp_message  = [];
    &set_initial_positions();
    $canvas->delete ($canvas->find ('all'));
    #Redraw the known servers at the top of the screen
    &restore_server_boxes();
    #Shrink the canvas back down.
    &resize_canvas();
}

sub draw_insert_break_dialog
{
    my $maximum_widget_width = 600;
    my $left_justify = 30;

    my $dialog = $window->Toplevel (width      => $left_justify * 2
				                + $maximum_widget_width,
				    background => $colors{window_background},
				    foreground => $colors{window_foreground},
				    takefocus  => 1);
    $dialog->title ('Insert a Visual Break');
    my $comment_entry = $dialog->Entry();
    my $comment_entry_label = $dialog->Label (text => 'Comment');
    my $OK_button = $dialog->Button (text => 'OK',
				     command => sub 
				                {
						    &insert_break ($comment_entry->get());
						    $dialog->withdraw();
						}
				     );
    my $Cancel_button = $dialog->Button (text => 'Cancel',
					 command => sub 
					            {
							$dialog->withdraw();
						    }
					 );
    $comment_entry_label->grid ($comment_entry);
    $OK_button->grid ($Cancel_button);
}

sub insert_break
{
    print "In callback \&insert_break...\n" if $debug;
    my $label = $_[0];

    my ($leftmost_x, $rightmost_x) = ($canvas->bbox ('all'))[0, 2];
    print "Drawing line from $leftmost_x to $rightmost_x\n" if $debug;
    
    if (defined $label)
    {
	print "Drawing text \"$label\" at ($leftmost_x, " . $arrow_vposition + 2 . ")\n" if $debug;

	$canvas->createText ($leftmost_x,
			     $arrow_vposition,
			     text    => $label,
			     font    => $settings{break_font},
			     #! width   => $settings{server_hsize},
			     fill    => $colors{break},
			     justify => 'right'
			     );
    }

    $arrow_vposition += $settings{break_offset_space};

    $canvas->createLine ($leftmost_x,
			 $arrow_vposition,
			 $rightmost_x,
			 $arrow_vposition,
			 width => $settings{break_width},
			 smooth => 1,
			 fill => $colors{break}
			 );
    &resize_canvas();
    $arrow_vposition += $settings{break_width} = $settings{break_offset_space};
}

#Convert Tk's description of a color mode to something more suitable for
#printing in a user interface.
sub canonical_to_verbose
{
    my $colormode = $_[0];

    if ($colormode eq 'color')
    {
	return 'Color';
    }
    elsif ($colormode eq 'gray')
    {
	return 'Greyscale';
    }
    elsif ($colormode eq 'mono')
    {
	return 'Black and White';
    }
    else
    {
	die "canonical_to_verbose was invoked with $colormode, which is not a valid colormode specification.";
    }
}

#Convert my user interface's description of a color mode to the canonical
#representation used by the Tk Canvas->postscript method.
sub verbose_to_canonical
{
    my $color_string = $_[0];
    
    if ($color_string eq 'Color')
    {
	return 'color';
    }
    elsif ($color_string eq 'Greyscale')
    {
	return 'gray';
    }
    elsif ($color_string eq 'Black and White')
    {
	return 'mono';
    }
    else
    {
	die "verbose_to_canonical was invoked with $color_string, which is not a valid color specification as displayed on the radiobuttons.";
    }
}

sub save_snapshot_preferences
{
    print "Saving snapshot preferences...\n" if $debug;

    open RC_FILE, ">$HOME/.tricorderrc"
	or die "Could not open $HOME/.tricorderrc for writing:  $!";
    print RC_FILE "#This is your personal configuration file for Tricorder,\n"
	         ."#the graphical call-flow diagram tool.  You probably shouldn't\n"
		 ."#edit it by hand; instead, run tricorder.pl and use the\n"
		 ."#'Snapshot Properties...' button to change your preferences graphically.\n\n";
    print RC_FILE Data::Dumper->Dump ([ \%snapshot_preferences ], [ '*snapshot_preferences' ]);
    close RC_FILE
	or &show_error ("Tricorder could not close $HOME/.tricorderrc; your preferences will not be saved.");

    print "Saved.\n" if $debug;
}

#Make sure there is space for all the servers.
sub readjust_canvas_width
{
    my $total_server_hspace = @server_list
	                      * ($settings{server_box_hsize}
				 + $settings{space_between_servers})
			      + $settings{server_start_hposition};
    if ($settings{canvas_width} < $total_server_hspace)
    {
	$settings{canvas_width} = $total_server_hspace;
    }

}

sub draw_server_box
{
    print "in draw_server_box\n" if $debug;
    my $server = $_[0];

    #! Later, make customizable through --naming-scheme option and through GUI
    my $server_name = $server->{logical_name}
                               || $server->{ip} . ':' . $server->{port};
    print "server $server_name is\n" . Data::Dumper->Dump ([ $server ]) if $debug_sip_messages;

    my $x1 = $server_box_hposition;
    my $y1 = $settings{server_vposition};
    my $x2 = $server_box_hposition + $settings{server_box_hsize};
    my $y2 = $y1 + $settings{server_box_vsize};

    print "Drawing box from ($x1, $y1) to ($x2, $y2) (for $server_name)\n" if $debug;
    my $box = $canvas->createRectangle
			       ($x1,
				$y1,
				$x2,
				$y2,
				fill    => $colors{server_box_fill},
				outline => $colors{server_box_outline},
				tags    => [ 'server_box' ]
				);
    
    if ($debug)
    {
	my @coords = $canvas->bbox (\$box);
	print "Drew box with bounding coordinates " . @coords;
    }
    my $horizontal_midpoint = $x1 + $settings{server_box_hsize} / 2;
    my $vertical_midpoint   = $y1 + $settings{server_box_vsize} / 2;

    #Record the horizontal coordinate of the _center_ of the server box
    #so draw_arrow can recall it later.
    $server->{hposition} = $horizontal_midpoint;

    $canvas->createText ($horizontal_midpoint,
			 $vertical_midpoint,
			 text    => $server_name,
			 #! width   => $settings{server_hsize},
			 fill    => $colors{server_box_outline},
			 font    => $settings{canvas_default_font},
			 justify => 'center',
			 tags    => 'server_label'
			 );

    my $bottom_of_the_canvas = $canvas->cget ('-height');
    $server_box_hposition += ($settings{server_box_hsize}
			      + $settings{space_between_servers});
}

sub draw_lines_under_server_boxes
{
    my $bottom_of_the_canvas = $_[0];

    map
    {
	$canvas->createLine ($_->{hposition},
			     $settings{server_vposition}
			     + $settings{server_box_vsize},
			     $_->{hposition},
			     $bottom_of_the_canvas,
			     fill   => $colors{server_box_outline},
			     smooth => 1,
			     tags   => [ 'vertical_line' ]
			     );
    } @server_list;
}

#! This function could be cleaned up somewhat by making %server_hash
#! into an object capable of returning false (i.e. '') when a value
#! can't be looked up.
sub draw_arrow
{
    my $message_number = $arrow_number;

    print "in draw_arrow\n" if $debug;

    my $sip_message = shift();

    #type-checking...
    ref $sip_message eq 'Telephony::SipMessage'
	or die "Error:  got a value of type " . ref ($sip_message)
	     . "; should have been of type Telephony::SipMessage\n";

    print "Drawing arrow for SIP message\n"
	. Data::Dumper->Dump ([ $sip_message ])
	    if $debug_sip_messages;
    
    my $source_ip   = $sip_message->{source_ip};
    my $source_port = $sip_message->{source_port};
    if (!defined $source_ip)
    {
	warn "draw_arrow got a message with no source IP; skipping...\n"
	   . "\tthe message was " . Data::Dumper->Dump ([ $sip_message ]) . "\n\n" if $debug_sip_messages;
	return;
    }
    if (!defined $source_port)
    {
	warn "draw_arrow got a message with no source port; skipping...\n"
	    . "\tthe message was " . Data::Dumper->Dump ([ $sip_message ]) . "\n\n" if $debug_sip_messages;
	return;
    }

    my $destination_ip   = $sip_message->{destination_ip};
    my $destination_port = $sip_message->{destination_port};
    if (!defined $destination_ip)
    {
	warn "draw_arrow got a message with no destination IP; skipping...\n"
	   . "\tthe message was " . Data::Dumper->Dump ([ $sip_message ]) . "\n\n" if $debug_sip_messages;
	return;
    }
    if (!defined $destination_port)
    {
	warn "draw_arrow got a message with no destination port; skipping...\n"
	   . "\tthe message was " . Data::Dumper->Dump ([ $sip_message ]) . "\n\n" if $debug_sip_messages;
	return;
    }

    my $source_ip_and_port      = $source_ip . ':' . $source_port;
    my $destination_ip_and_port = $destination_ip . ':' . $destination_port;
    
    my $source = $server_hash{$source_ip_and_port};
    print "Got corresponding server:  " . Data::Dumper->Dump ([ $source ]) . "\n" if $debug_sip_messages;
    my $source_hposition      = (ref $source eq 'Telephony::SipServer') ?
	                        $source->{hposition} :
				0;

    my $destination = $server_hash{$destination_ip_and_port};
    print "Got corresponding server:  " . Data::Dumper->Dump ([ $destination ]) . "\n" if $debug_sip_messages;
    my $destination_hposition = (ref $destination eq 'Telephony::SipServer') ?    
	                        $destination->{hposition} :
                                0;    

    print "Drawing arrow for source $source_ip_and_port and destination $destination_ip_and_port...\n" if $debug_sip_messages;

    my $x1 = $source_hposition;
    my $x2 = $destination_hposition;
    my $y1 = $arrow_vposition;
    my $y2 = $arrow_vposition;

    my $color_coding = &get_arrow_color ($sip_message->{reqres});
    print "Drawing $color_coding arrow from ($x1, $y1) to ($x2, $y2)\n" if $debug;
    $canvas->createLine ($x1,
			 $y1,
			 $x2,
			 $y2,
			 arrow  => 'last',
			 smooth => 1,
			 fill   => $color_coding,
			 tags => [ "message_arrow_$message_number" ]
			 );

    my $label = $canvas->createText ($x1 < $x2 ? $x1 : $x2,
				     $y1 - 12,
				     text    => "$message_number. $sip_message->{reqres}",
				     anchor => 'nw',
				     fill    => $color_coding,
				     font    => $settings{canvas_default_font},
				     tags    => [ "message_label_$message_number" ]
				     );
    $canvas->bind ($label,
		   '<Button-1>',
		   sub
		   {
		       print "Before call to \&draw_header_balloons, \$message_number is $message_number\n" if $debug;
		       &draw_header_balloons ($message_number,
					      $sip_message, #closure!
					      $x1 < $x2 ? $x1 : $x2,
					      $y1);
		   }
                  );

    #Move down the screen...
    $arrow_vposition += $settings{space_between_arrows};
    $arrow_number++;
}

sub get_arrow_color
{
    my $color;

    my $reqres = shift();
    print "Getting color" if $debug;
    print " for $reqres arrow..." if $debug_sip_messages;
    print "\n" if $debug;

    #Response codes:  Match on the first digit (which indicates the
    #general type of response)
    if ($reqres =~ /^(\d)/)
    {
	$color = $colors{(grep /$1/, (keys %colors))[0]};
    }
    #Requests:  Match on the whole name
    else
    {
	$color = $colors{$reqres};
    }

    #Set a default color for unknown reqres's.
    if (!defined $color)
    {
	$color = 'black';
    }
    print "its color is $color\n" if $debug;
    return $color;
}

sub draw_header_balloons
{
    my ($message_number, $message, $x, $y) = @_; #message is guaranteed to be a
                                              #Telephony::SipMessage by earlier
                                              #type-checking

    print "After call to \&draw_header_balloons, \$message_number is $message_number\n" if $debug;

    defined $message->{sip_headers}
        or return;

    my $sip_balloon = FancyBalloon->new ($debug_sip_messages,
					 $message->{sip_headers},
					 $colors{sip_header_name},
					 $colors{sip_header_value},
					 $canvas);
    if (! defined $sip_balloon)
    {
	warn "FancyBalloon->new returned undef; there is no SIP balloon to display.";
	return;
    }
    
    print "Constructed a SIP FancyBalloon:\n" . Data::Dumper->Dump ([ $sip_balloon ]) if $debug;

    $sip_balloon->title ("$message_number. $message->{reqres}");
    #! $sip_balloon->place (-x => $x, -y => $y + 5); #in $canvas, its parent
    #! Ouch, that doesn't work.  I'll have to fiddle with the window manager!
    if (defined $message->{sdp_headers})
    {
	my $sdp_balloon = FancyBalloon->new ($debug_sip_messages,
					     $message->{sdp_headers},
	                                     $colors{sdp_header_name},
	                                     $colors{sdp_header_value},
					     $canvas);

        if (! defined $sdp_balloon)
	{
	    warn "FancyBalloon->new returned undef; there is no SDP balloon to display.";
	    return;
	}

        print "Constructed an SDP FancyBalloon:" . Data::Dumper->Dump ([ $sdp_balloon ]) if $debug;
        $sdp_balloon->title ("$message_number. $message->{reqres}");
        #! $sdp_balloon->place (-x => $x, -y => $y + 5 + $sip_balloon->height()); #in $canvas, its parent
	#! Ouch, that doesn't work.  I'll have to fiddle with the window manager!
    }
}

sub get_quip
#Return a random Star Trek quotation for the user's enjoyment
{
    my @quips = ("Cap'n, I'm giving 'er all I've got!",
		 "It's life, Jim, but not as we know it.",
		 "He's dead, Jim.",
		 "Beam me up, Scotty.",
		 "Live long and prosper.",
		 );
    srand();
    return $quips[rand ($#quips + 1)];
}

#Lecture the user on proper syntax.
sub usage
{
    my ($message, $exit_value) = @_;
    print STDERR $message if defined $message;

    print STDERR <<fnord;
usage:  $program_name --help
        $program_name [--log-file log_file_name] OPTIONS
	$program_name [--tee output_file_path] [--ipgrab-expression expr] OPTIONS
	where options are of the form:
        [--silent | --debug [--debug-sip-messages]]
	[--cisco]
	[--heuristic assume-server
                   | use-vovida-ip-numbering-scheme
                   | use-sip-hints]
	[--paginate-snapshots [dumb | smart]]


        where
        
        --help prints this message,

	--log-file specifies an ipgrab log* to examine in graphical mode.
	  If --log-file is not specified, $program_name starts an ipgrab
	  process of its own and reads from it, dynamically updating the
          display as input becomes available.

        --tee tells $program_name to write the output from such an
          ipgrab subprocess to the file named output_file_path,
	  so that the raw header information is not lost.  (To tee to
          standard output or standard error, use /dev/stdout or /dev/stderr
	  respectively.)

	--ipgrab-expression specifies the expression to run with ipgrab,
          e.g. port 5060 or port 5070 or port 5074.  See the ipgrab(8)
	  man page for more details.

	--cisco tells $program_name to massage each messsage's
          source and destination IP to account for the fact that Cisco IP phones
          use high local ports, rather than 5060.  (Otherwise the phone appears as
	  two different boxes, since a box is determined by the combination of
          IP and port.)

        --silent tells $program_name not to print messages about
          its activity to standard output.  (The default is to print
          some messages),

	--debug tells $program_name to print basic debugging messages,
          useful for $program_name developers,

	--debug-sip-messages tells $program_name to print detailed
	  debugging messages while parsing the SIP transaction,
          This slows down $program_name launch significantly, since the
	  parser gives verbose debugging output for each SIP message.
	  Specifying --debug-sip-messages also turns on normal debugging
	  messages, since it is hard to follow the flow of control without
	  them.

	--heuristic tells $program_name to discriminate between servers and
          endpoints using one of the following heuristics:

               assume-server:  Treat everything as a server.  This produces
               perfectly usable diagrams for scenarios with a small number
               of endpoints, but it breaks down at some point, since each
               server is represented as a separate box in the diagram.

               use-vovida-ip-numbering-scheme:  Examine the last octet
               of the message sender and recipients' IPs to decide if they
	       are servers or endpoints.  In the Vovida scheme, .180,
	       .200, .110, and .210 are reserved for servers.  This method
	       is fast, but, of course, it relies on the local setup using
	       those numbers.

               use-sip-hints:  Examine the actual SIP messages.  This is
               the most costly method, but also the most reliable.

	  The default is assume-server.

	  --paginate-snapshots tells $program_name to separate its Postscript
            snapshots (taken when you press the "Take a Snapshot!" button)
	    into files of a certain size.  The argument dumb means to
	    separate strictly based on page size, without concern for
	    orphaned or dangling messages or servers; smart means to
            take message and server boundaries into account as well.
            Hence smart is liable to increase the number of pages,
            as trailing data that do not fit semantically on a given
            page are relegated to its successor.  Note also that smart
	    scales the snapshot horizontally so that it occupies only
            one page horizontally.  Otherwise, there is no way to guarantee
            that long message arrows will fit in the horizontal page length
            guaranteed by the user.

            If no argument is given, $program_name assumes dumb.

            The size of pages is configurable through the GUI, via
            the "Snapshot Properties..." button.

*For the log file to work, it must have been produced by ipgrab using the
-t, -n, and -l options.  $program_name currently does not
understand more verbose ipgrab output.

fnord

exit $exit_value || 1;
}
