#if !defined(POLL_DOT_H)
#define POLL_DOT_H

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const Poll_hxx_Version = 
"$Id: Poll.hxx,v 1.12 2000/08/11 00:32:59 bogawa Exp $";

#include <hash_map>
#include <vector>
#include <sys/poll.h>
#include "Sptr.hxx"
#include "UDPSocket.hxx"
#include "VMutex.h"
#include "Writer.hxx"

namespace Vovida
{

class Protocol;
class TransportException;
class ProtocolException;
class ConnectionBrokenException;
class UDPSocket;
struct pollfdHash;
struct pollfdEqual;

// Needed for hash_map of pollfds
//
struct pollfdHash
{
    size_t operator()(const pollfd & src) const;
};

// Needed for hash_map of pollfds
//
struct pollfdEqual
{
    bool operator()(const pollfd & left, const pollfd & right) const;
};

/// A Poll wraps the operating system's poll() call.
///
/// The Poll and Protocol classes together form a type of Subject/Observer
/// (see Design Patterns, Gamma, et al. page 293) pattern. The Poll object
/// is the subject whereas the Protocol is the Observer. The Protocol object
/// register's itself with the Poll. The protocol is interested in being 
/// notified when it's file descriptor is active. Thus the observer provides 
/// and an aspect to the server to reduce messaging activity. As well, since 
/// the observer provides an aspect, the push model is used.
///
class Poll : public Writer
{
    public:
    	    	    	    	    Poll();
    	virtual     	    	    ~Poll();

    	    	/// Register the protocol's file descriptor for
		/// activity detection. 
		///	
	    	void	    	    registerProtocol(Protocol &);
    	    	void	    	    updateProtocol(Protocol &);
	    	void	    	    unregisterProtocol(Protocol &);

    	    	/// A blocking call waiting for activity on the
		/// file descriptors of the registered protocols.
		/// If a signal interrupts the poll, 0 is returned,
		/// otherwise the number of active file descriptors is
		/// returned.
		///
    	    	int 	    	    poll(int timeout = -1)
		    	    	    	throw ( TransportException );

    	    	/// process the activity on the given protocols.
		///
		void 	    	    processProtocols(int numberFdsActive)
    	    	    	    	    	throw ( ProtocolException );
		
    	    	/// interrupt() should only be called from a thread
		/// different from the thread the majority of the
		/// work for this object is being done.
		///		
		void	    	    interrupt()
		    	    	    	throw ( TransportException );
					
		/// interruptCount returns the number of outstanding 
		/// interrupts pending on the poll, if any.
		///
		u_int8_t    	    interruptCount() const;
		

	virtual ostream &           writeTo(ostream &) const;
					
    private:

    	    	void	    	    addInterruptorToFds();
		void	    	    addProtocolsToFds();

    	    	bool	    	    processInterruptor(pollfd &);
    	    	bool	    	    processPriority(pollfd &, Protocol *);
    	    	bool	    	    processIncoming(pollfd &, Protocol *);
    	    	bool	    	    processOutgoing(pollfd &, Protocol *);
    	    	bool	    	    processHangup(
		    	    	    	pollfd &, 
					Protocol *,
					ConnectionBrokenException * = 0
				    );
    	    	bool	    	    processError(
		    	    	    	pollfd &, 
					Protocol *,
					TransportException * = 0
					);
				    
     	    	typedef hash_map<pollfd, 
 		    	    	 Protocol *, 
 		    	    	 pollfdHash,
 				 pollfdEqual> 	ProtocolMap;

		ProtocolMap 	    protocols_;
	    	bool	    	    protocolsChanged_;
	
    	    	vector<pollfd>	    fds_;
	
    	    	Sptr<UDPSocket>	    interruptor_;
		u_int8_t   	    interruptCount_;
	    	VMutex	    	    interruptorMutex_;
	
    		// Suppress copying
		//
    		Poll(const Poll &);
		Poll & operator=(const Poll &);
};

} // namespace Vovida

#endif // !defined(POLL_DOT_H)
