/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const StateInterface_cxx_Version = 
"$Id: StateInterface.cxx,v 1.1 2000/09/15 02:34:31 icahoon Exp $";


#include <cassert>
#include "StateInterface.hxx"
#include "Stimulus.hxx"
#include "Action.hxx"
#include "InvalidStimulus.hxx"


using namespace Vovida;


StateInterface::StateInterface(
    Action  	    *	onEntry,
    Action  	    *	onExit
)
    :	entry_(onEntry),
    	exit_(onExit)
{
}


StateInterface::~StateInterface()
{
}


void		    
StateInterface::addAction(
    const Stimulus 	    &	stimulus,
    Action  	    	    &	action,
    StateInterface          &	nextState
)
{
    bool rc =	stimulusTable_.insert(
    	    	    StimulusPair(stimulus, StimulusData(action, nextState))
	      	).second;

    assert( rc );
}


void        	    
StateInterface::enter() throw ()
{
    if ( entry_ )
    {
    	entry_->action();
    }
}


StateInterface &    
StateInterface::stimulate(
    const Stimulus  &   stimulus
) throw ()
{

    // Lookup the stimulus in the stimulus table. 
    //
    StimulusTable::iterator it = stimulusTable_.find(stimulus);
    
    // If it doesn't exist, throw InvalidStimulus. 
    //
    if ( it == stimulusTable_.end() )
    {
    	throw InvalidStimulus(stimulus);
    }
    
    // If the stimulus exists first perform the exit action on this state, 
    // then perform the action associated with the stimulus, and finally
    // perform the entry action on the next state.
    //
    // Note that the exit action, the action associated with this stimulus, 
    // and the entry action of the next state may throw an exception.
    //
    exit();
    
    (*it).second.action_.action(stimulus);
    
    StateInterface &	nextState = (*it).second.nextState_;
    
    nextState.enter();
    
    return ( nextState );
}


void        	    
StateInterface::exit() throw ()
{
    if ( exit_ )
    {
    	exit_->action();
    }
}


ostream &           
StateInterface::writeTo(ostream & out) const
{
    out << name() 
    	<< "\n\tentry: " << ( entry_ ? *entry_ : Action() )
	<< "\n\texit:  " << ( exit_  ? *exit_  : Action() )
	<< "\n\tstimuli: {";
    
    for (   StimulusTable::const_iterator it = stimulusTable_.begin();
    	    it != stimulusTable_.end();
	    it++
	)
    {
    	out << "\n\t\tstimulus:" << (*it).first
	    << ",\taction: " << (*it).second.action_
	    << ",\tnextState: " << (*it).second.nextState_.name();
    }
    
    return ( out << "\n}\n" );
}


const char *	    
StateInterface::name() const
{
    return ( "NullState" );
}


StateInterface::StimulusData::StimulusData(
    Action  	    	&   action,
    StateInterface      &   nextState
)   
    :   action_(action),
    	nextState_(nextState)
{
}


